@@ -14,7 +14,8 @@
 #
 from __future__ import annotations
 
-from typing import Any, Sequence
+from collections.abc import Sequence
+from typing import Any
 
 from . import Image
 
